@@ -18,7 +18,7 @@
  */
 
 /***************************************************************************
- * Copyright (C) 2013-2016 Ping Identity Corporation
+ * Copyright (C) 2013-2017 Ping Identity Corporation
  * All rights reserved.
  *
  * For further information please contact:
@@ -45,7 +45,7 @@
  * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- * @Author: Hans Zandbelt - hzandbelt@pingidentity.com
+ * @Author: Hans Zandbelt - hans.zandbelt@zmartzone.eu
  */
 
 #include <apr_strings.h>
@@ -105,12 +105,11 @@ int oidc_base64url_encode(request_rec *r, char **dst, const char *src,
 /*
  * base64url decode a string
  */
-int oidc_base64url_decode(request_rec *r, char **dst, const char *src) {
+int oidc_base64url_decode(apr_pool_t *pool, char **dst, const char *src) {
 	if (src == NULL) {
-		oidc_error(r, "not decoding anything; src=NULL");
 		return -1;
 	}
-	char *dec = apr_pstrdup(r->pool, src);
+	char *dec = apr_pstrdup(pool, src);
 	int i = 0;
 	while (dec[i] != '\0') {
 		if (dec[i] == '-')
@@ -125,56 +124,120 @@ int oidc_base64url_decode(request_rec *r, char **dst, const char *src) {
 	case 0:
 		break;
 	case 2:
-		dec = apr_pstrcat(r->pool, dec, "==", NULL);
+		dec = apr_pstrcat(pool, dec, "==", NULL);
 		break;
 	case 3:
-		dec = apr_pstrcat(r->pool, dec, "=", NULL);
+		dec = apr_pstrcat(pool, dec, "=", NULL);
 		break;
 	default:
 		return 0;
 	}
 	int dlen = apr_base64_decode_len(dec);
-	*dst = apr_palloc(r->pool, dlen);
+	*dst = apr_palloc(pool, dlen);
 	return apr_base64_decode(*dst, dec);
 }
 
-/*
- * encrypt and base64url encode a string
- */
-int oidc_encrypt_base64url_encode_string(request_rec *r, char **dst,
-		const char *src) {
-	oidc_cfg *c = ap_get_module_config(r->server->module_config,
-			&auth_openidc_module);
-	int crypted_len = strlen(src) + 1;
-	unsigned char *crypted = oidc_crypto_aes_encrypt(r, c,
-			(unsigned char *) src, &crypted_len);
-	if (crypted == NULL) {
-		oidc_error(r, "oidc_crypto_aes_encrypt failed");
-		return -1;
+apr_byte_t oidc_util_jwt_create(request_rec *r, const char *secret,
+		json_t *payload, char **compact_encoded_jwt) {
+
+	apr_byte_t rv = FALSE;
+	oidc_jose_error_t err;
+
+	oidc_jwk_t *jwk = NULL;
+	oidc_jwt_t *jwt = NULL;
+	oidc_jwt_t *jwe = NULL;
+
+	if (oidc_util_create_symmetric_key(r, secret, 0, "sha256", FALSE,
+			&jwk) == FALSE)
+		goto end;
+
+	jwt = oidc_jwt_new(r->pool, TRUE, FALSE);
+	if (jwt == NULL) {
+		oidc_error(r, "creating JWT failed");
+		goto end;
 	}
-	return oidc_base64url_encode(r, dst, (const char *) crypted, crypted_len, 1);
+
+	jwt->header.alg = apr_pstrdup(r->pool, "HS256");
+	jwt->payload.value.json = payload;
+
+	if (oidc_jwt_sign(r->pool, jwt, jwk, &err) == FALSE) {
+		oidc_error(r, "signing JWT failed: %s", oidc_jose_e2s(r->pool, err));
+		goto end;
+	}
+
+	jwe = oidc_jwt_new(r->pool, TRUE, FALSE);
+	if (jwe == NULL) {
+		oidc_error(r, "creating JWE failed");
+		goto end;
+	}
+
+	jwe->header.alg = "dir";
+	jwe->header.enc = "A256GCM";
+
+	const char *cser = oidc_jwt_serialize(r->pool, jwt, &err);
+	if (oidc_jwt_encrypt(r->pool, jwe, jwk, cser, compact_encoded_jwt,
+			&err) == FALSE) {
+		oidc_error(r, "encrypting JWT failed: %s", oidc_jose_e2s(r->pool, err));
+		goto end;
+	}
+
+	rv = TRUE;
+
+	end:
+
+	if (jwe != NULL)
+		oidc_jwt_destroy(jwe);
+	if (jwk != NULL)
+		oidc_jwk_destroy(jwk);
+	if (jwt != NULL) {
+		jwt->payload.value.json = NULL;
+		oidc_jwt_destroy(jwt);
+	}
+
+	return rv;
 }
 
-/*
- * decrypt and base64url decode a string
- */
-int oidc_base64url_decode_decrypt_string(request_rec *r, char **dst,
-		const char *src) {
-	oidc_cfg *c = ap_get_module_config(r->server->module_config,
-			&auth_openidc_module);
-	char *decbuf = NULL;
-	int dec_len = oidc_base64url_decode(r, &decbuf, src);
-	if (dec_len <= 0) {
-		oidc_error(r, "oidc_base64url_decode failed");
-		return -1;
+apr_byte_t oidc_util_jwt_verify(request_rec *r, const char *secret,
+		const char *compact_encoded_jwt, json_t **result) {
+
+	oidc_debug(r, "enter: JWT header=%s",
+			oidc_proto_peek_jwt_header(r, compact_encoded_jwt, NULL));
+
+	apr_byte_t rv = FALSE;
+	oidc_jose_error_t err;
+
+	oidc_jwk_t *jwk = NULL;
+	oidc_jwt_t *jwt = NULL;
+
+	if (oidc_util_create_symmetric_key(r, secret, 0, "sha256", FALSE,
+			&jwk) == FALSE)
+		goto end;
+
+	apr_hash_t *keys = apr_hash_make(r->pool);
+	apr_hash_set(keys, "", APR_HASH_KEY_STRING, jwk);
+
+	if (oidc_jwt_parse(r->pool, compact_encoded_jwt, &jwt, keys, &err) == FALSE) {
+		oidc_error(r, "parsing JWT failed: %s", oidc_jose_e2s(r->pool, err));
+		goto end;
 	}
-	*dst = (char *) oidc_crypto_aes_decrypt(r, c, (unsigned char *) decbuf,
-			&dec_len);
-	if (*dst == NULL) {
-		oidc_error(r, "oidc_crypto_aes_decrypt failed");
-		return -1;
+
+	if (oidc_jwt_verify(r->pool, jwt, keys, &err) == FALSE) {
+		oidc_error(r, "verifying JWT failed: %s", oidc_jose_e2s(r->pool, err));
+		goto end;
 	}
-	return dec_len;
+
+	*result = json_deep_copy(jwt->payload.value.json);
+
+	rv = TRUE;
+
+	end:
+
+	if (jwk != NULL)
+		oidc_jwk_destroy(jwk);
+	if (jwt != NULL)
+		oidc_jwt_destroy(jwt);
+
+	return rv;
 }
 
 /*
@@ -269,9 +332,10 @@ char *oidc_util_unescape_string(const request_rec *r, const char *str) {
  * HTML escape a string
  */
 char *oidc_util_html_escape(apr_pool_t *pool, const char *s) {
+	// TODO: this has performance/memory issues for large chunks of HTML
 	const char chars[6] = { '&', '\'', '\"', '>', '<', '\0' };
 	const char * const replace[] =
-			{ "&amp;", "&apos;", "&quot;", "&gt;", "&lt;", };
+	{ "&amp;", "&apos;", "&quot;", "&gt;", "&lt;", };
 	unsigned int i, j = 0, k, n = 0, len = strlen(chars);
 	int m = 0;
 	char *r = apr_pcalloc(pool, strlen(s) * 6);
@@ -385,28 +449,48 @@ char *oidc_get_current_url(request_rec *r) {
 	return url;
 }
 
-/* maximum size of any response returned in HTTP calls */
-#define OIDC_CURL_MAX_RESPONSE_SIZE 65536
-
 /* buffer to hold HTTP call responses */
 typedef struct oidc_curl_buffer {
-	char buf[OIDC_CURL_MAX_RESPONSE_SIZE];
-	size_t written;
+	request_rec *r;
+	char *memory;
+	size_t size;
 } oidc_curl_buffer;
 
+/* maximum acceptable size of HTTP responses: 1 Mb */
+#define OIDC_CURL_MAX_RESPONSE_SIZE 1024 * 1024
+
 /*
  * callback for CURL to write bytes that come back from an HTTP call
  */
-size_t oidc_curl_write(const void *ptr, size_t size, size_t nmemb, void *stream) {
-	oidc_curl_buffer *curlBuffer = (oidc_curl_buffer *) stream;
+size_t oidc_curl_write(void *contents, size_t size, size_t nmemb, void *userp) {
+	size_t realsize = size * nmemb;
+	oidc_curl_buffer *mem = (oidc_curl_buffer *) userp;
 
-	if ((nmemb * size) + curlBuffer->written >= OIDC_CURL_MAX_RESPONSE_SIZE)
+	/* check if we don't run over the maximum buffer/memory size for HTTP responses */
+	if (mem->size + realsize > OIDC_CURL_MAX_RESPONSE_SIZE) {
+		oidc_error(mem->r,
+				"HTTP response larger than maximum allowed size: current size=%ld, additional size=%ld, max=%d",
+				mem->size, realsize, OIDC_CURL_MAX_RESPONSE_SIZE);
 		return 0;
+	}
 
-	memcpy((curlBuffer->buf + curlBuffer->written), ptr, (nmemb * size));
-	curlBuffer->written += (nmemb * size);
+	/* allocate the new buffer for the current + new response bytes */
+	char *newptr = apr_palloc(mem->r->pool, mem->size + realsize + 1);
+	if (newptr == NULL) {
+		oidc_error(mem->r,
+				"memory allocation for new buffer of %ld bytes failed",
+				mem->size + realsize + 1);
+		return 0;
+	}
 
-	return (nmemb * size);
+	/* copy over the data from current memory plus the cURL buffer */
+	memcpy(newptr, mem->memory, mem->size);
+	memcpy(&(newptr[mem->size]), contents, realsize);
+	mem->size += realsize;
+	mem->memory = newptr;
+	mem->memory[mem->size] = 0;
+
+	return realsize;
 }
 
 /* context structure for encoding parameters */
@@ -435,7 +519,8 @@ static apr_byte_t oidc_util_http_call(request_rec *r, const char *url,
 		const char *data, const char *content_type, const char *basic_auth,
 		const char *bearer_token, int ssl_validate_server,
 		const char **response, int timeout, const char *outgoing_proxy,
-		apr_array_header_t *pass_cookies) {
+		apr_array_header_t *pass_cookies, const char *ssl_cert,
+		const char *ssl_key) {
 	char curlError[CURL_ERROR_SIZE];
 	oidc_curl_buffer curlBuffer;
 	CURL *curl;
@@ -454,6 +539,9 @@ static apr_byte_t oidc_util_http_call(request_rec *r, const char *url,
 		return FALSE;
 	}
 
+	/* set the error buffer as empty before performing a request */
+	curlError[0] = 0;
+
 	/* some of these are not really required */
 	curl_easy_setopt(curl, CURLOPT_HEADER, 0L);
 	curl_easy_setopt(curl, CURLOPT_NOPROGRESS, 1L);
@@ -466,10 +554,11 @@ static apr_byte_t oidc_util_http_call(request_rec *r, const char *url,
 	curl_easy_setopt(curl, CURLOPT_TIMEOUT, timeout);
 
 	/* setup the buffer where the response will be written to */
-	curlBuffer.written = 0;
-	memset(curlBuffer.buf, '\0', sizeof(curlBuffer.buf));
-	curl_easy_setopt(curl, CURLOPT_WRITEDATA, &curlBuffer);
+	curlBuffer.r = r;
+	curlBuffer.memory = NULL;
+	curlBuffer.size = 0;
 	curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, oidc_curl_write);
+	curl_easy_setopt(curl, CURLOPT_WRITEDATA, (void * )&curlBuffer);
 
 #ifndef LIBCURL_NO_CURLPROTO
 	curl_easy_setopt(curl, CURLOPT_REDIR_PROTOCOLS,
@@ -517,6 +606,11 @@ static apr_byte_t oidc_util_http_call(request_rec *r, const char *url,
 		curl_easy_setopt(curl, CURLOPT_USERPWD, basic_auth);
 	}
 
+	if (ssl_cert != NULL)
+		curl_easy_setopt(curl, CURLOPT_SSLCERT, ssl_cert);
+	if (ssl_key != NULL)
+		curl_easy_setopt(curl, CURLOPT_SSLKEY, ssl_key);
+
 	if (data != NULL) {
 		/* set POST data */
 		curl_easy_setopt(curl, CURLOPT_POSTFIELDS, data);
@@ -534,26 +628,29 @@ static apr_byte_t oidc_util_http_call(request_rec *r, const char *url,
 	if (h_list != NULL)
 		curl_easy_setopt(curl, CURLOPT_HTTPHEADER, h_list);
 
-	/* gather cookies that we need to pass on from the incoming request */
-	char *cookie_string = NULL;
-	for (i = 0; i < pass_cookies->nelts; i++) {
-		const char *cookie_name = ((const char**) pass_cookies->elts)[i];
-		char *cookie_value = oidc_util_get_cookie(r, cookie_name);
-		if (cookie_value != NULL) {
-			cookie_string =
-					(cookie_string == NULL) ?
-							apr_psprintf(r->pool, "%s=%s", cookie_name,
-									cookie_value) :
-									apr_psprintf(r->pool, "%s; %s=%s", cookie_string,
-											cookie_name, cookie_value);
+	if (pass_cookies != NULL) {
+		/* gather cookies that we need to pass on from the incoming request */
+		char *cookie_string = NULL;
+		for (i = 0; i < pass_cookies->nelts; i++) {
+			const char *cookie_name = ((const char**) pass_cookies->elts)[i];
+			char *cookie_value = oidc_util_get_cookie(r, cookie_name);
+			if (cookie_value != NULL) {
+				cookie_string =
+						(cookie_string == NULL) ?
+								apr_psprintf(r->pool, "%s=%s", cookie_name,
+										cookie_value) :
+										apr_psprintf(r->pool, "%s; %s=%s",
+												cookie_string, cookie_name,
+												cookie_value);
+			}
 		}
-	}
 
-	/* see if we need to pass any cookies */
-	if (cookie_string != NULL) {
-		oidc_debug(r, "passing browser cookies on backend call: %s",
-				cookie_string);
-		curl_easy_setopt(curl, CURLOPT_COOKIE, cookie_string);
+		/* see if we need to pass any cookies */
+		if (cookie_string != NULL) {
+			oidc_debug(r, "passing browser cookies on backend call: %s",
+					cookie_string);
+			curl_easy_setopt(curl, CURLOPT_COOKIE, cookie_string);
+		}
 	}
 
 	/* set the target URL */
@@ -562,15 +659,20 @@ static apr_byte_t oidc_util_http_call(request_rec *r, const char *url,
 	/* call it and record the result */
 	int rv = TRUE;
 	if (curl_easy_perform(curl) != CURLE_OK) {
-		oidc_error(r, "curl_easy_perform() failed on: %s (%s)", url, curlError);
+		oidc_error(r, "curl_easy_perform() failed on: %s (%s)", url,
+				curlError[0] ? curlError : "");
 		rv = FALSE;
 		goto out;
 	}
 
-	*response = apr_pstrndup(r->pool, curlBuffer.buf, curlBuffer.written);
+	long response_code;
+	curl_easy_getinfo(curl, CURLINFO_RESPONSE_CODE, &response_code);
+	oidc_debug(r, "HTTP response code=%ld", response_code);
+
+	*response = apr_pstrndup(r->pool, curlBuffer.memory, curlBuffer.size);
 
 	/* set and log the response */
-	oidc_debug(r, "response=%s", *response);
+	oidc_debug(r, "response=%s", *response ? *response : "");
 
 out:
 
@@ -589,7 +691,8 @@ apr_byte_t oidc_util_http_get(request_rec *r, const char *url,
 		const apr_table_t *params, const char *basic_auth,
 		const char *bearer_token, int ssl_validate_server,
 		const char **response, int timeout, const char *outgoing_proxy,
-		apr_array_header_t *pass_cookies) {
+		apr_array_header_t *pass_cookies, const char *ssl_cert,
+		const char *ssl_key) {
 
 	if ((params != NULL) && (apr_table_elts(params)->nelts > 0)) {
 		oidc_http_encode_t data = { r, "" };
@@ -601,7 +704,7 @@ apr_byte_t oidc_util_http_get(request_rec *r, const char *url,
 
 	return oidc_util_http_call(r, url, NULL, NULL, basic_auth, bearer_token,
 			ssl_validate_server, response, timeout, outgoing_proxy,
-			pass_cookies);
+			pass_cookies, ssl_cert, ssl_key);
 }
 
 /*
@@ -611,7 +714,8 @@ apr_byte_t oidc_util_http_post_form(request_rec *r, const char *url,
 		const apr_table_t *params, const char *basic_auth,
 		const char *bearer_token, int ssl_validate_server,
 		const char **response, int timeout, const char *outgoing_proxy,
-		apr_array_header_t *pass_cookies) {
+		apr_array_header_t *pass_cookies, const char *ssl_cert,
+		const char *ssl_key) {
 
 	const char *data = NULL;
 	if ((params != NULL) && (apr_table_elts(params)->nelts > 0)) {
@@ -625,7 +729,7 @@ apr_byte_t oidc_util_http_post_form(request_rec *r, const char *url,
 	return oidc_util_http_call(r, url, data,
 			"application/x-www-form-urlencoded", basic_auth, bearer_token,
 			ssl_validate_server, response, timeout, outgoing_proxy,
-			pass_cookies);
+			pass_cookies, ssl_cert, ssl_key);
 }
 
 /*
@@ -634,7 +738,8 @@ apr_byte_t oidc_util_http_post_form(request_rec *r, const char *url,
 apr_byte_t oidc_util_http_post_json(request_rec *r, const char *url,
 		const json_t *json, const char *basic_auth, const char *bearer_token,
 		int ssl_validate_server, const char **response, int timeout,
-		const char *outgoing_proxy, apr_array_header_t *pass_cookies) {
+		const char *outgoing_proxy, apr_array_header_t *pass_cookies,
+		const char *ssl_cert, const char *ssl_key) {
 
 	char *data = NULL;
 	if (json != NULL) {
@@ -645,7 +750,7 @@ apr_byte_t oidc_util_http_post_json(request_rec *r, const char *url,
 
 	return oidc_util_http_call(r, url, data, "application/json", basic_auth,
 			bearer_token, ssl_validate_server, response, timeout,
-			outgoing_proxy, pass_cookies);
+			outgoing_proxy, pass_cookies, ssl_cert, ssl_key);
 }
 
 /*
@@ -668,15 +773,14 @@ static char *oidc_util_get_path(request_rec *r) {
  */
 static char *oidc_util_get_cookie_path(request_rec *r) {
 	char *rv = NULL, *requestPath = oidc_util_get_path(r);
-	oidc_dir_cfg *d = ap_get_module_config(r->per_dir_config,
-			&auth_openidc_module);
-	if (d->cookie_path != NULL) {
-		if (strncmp(d->cookie_path, requestPath, strlen(d->cookie_path)) == 0)
-			rv = d->cookie_path;
+	char *cookie_path = oidc_cfg_dir_cookie_path(r);
+	if (cookie_path != NULL) {
+		if (strncmp(cookie_path, requestPath, strlen(cookie_path)) == 0)
+			rv = cookie_path;
 		else {
 			oidc_warn(r,
 					"OIDCCookiePath (%s) not a substring of request path, using request path (%s) for cookie",
-					d->cookie_path, requestPath);
+					cookie_path, requestPath);
 			rv = requestPath;
 		}
 	} else {
@@ -693,7 +797,7 @@ void oidc_util_set_cookie(request_rec *r, const char *cookieName,
 
 	oidc_cfg *c = ap_get_module_config(r->server->module_config,
 			&auth_openidc_module);
-	char *headerString, *currentCookies, *expiresString = NULL;
+	char *headerString, *expiresString = NULL;
 
 	/* see if we need to clear the cookie */
 	if (apr_strnatcmp(cookieValue, "") == 0)
@@ -709,34 +813,25 @@ void oidc_util_set_cookie(request_rec *r, const char *cookieName,
 
 	/* construct the cookie value */
 	headerString = apr_psprintf(r->pool, "%s=%s;Path=%s%s%s%s%s", cookieName,
-			cookieValue,
-			oidc_util_get_cookie_path(r),
+			cookieValue, oidc_util_get_cookie_path(r),
 			(expiresString == NULL) ?
 					"" : apr_psprintf(r->pool, "; expires=%s", expiresString),
-			c->cookie_domain != NULL ?
-					apr_psprintf(r->pool, ";Domain=%s", c->cookie_domain) : "",
-			((apr_strnatcasecmp("https", oidc_get_current_url_scheme(r)) == 0) ?
-					";Secure" : ""),
-			c->cookie_http_only != FALSE ? ";HttpOnly" : "");
+					c->cookie_domain != NULL ?
+							apr_psprintf(r->pool, ";Domain=%s", c->cookie_domain) : "",
+							((apr_strnatcasecmp("https", oidc_get_current_url_scheme(r)) == 0) ?
+									";Secure" : ""),
+									c->cookie_http_only != FALSE ? ";HttpOnly" : "");
 
 	/* sanity check on overall cookie value size */
 	if (strlen(headerString) > 4093) {
 		oidc_warn(r,
 				"the length of the cookie value (%lu) is greater than 4093(!) bytes, this may not work with all browsers/server combinations: consider switching to a server side caching!",
-				(unsigned long)strlen(headerString));
+				(unsigned long )strlen(headerString));
 	}
 
 	/* use r->err_headers_out so we always print our headers (even on 302 redirect) - headers_out only prints on 2xx responses */
 	apr_table_add(r->err_headers_out, "Set-Cookie", headerString);
 
-	/* see if we need to add it to existing cookies */
-	if ((currentCookies = (char *) apr_table_get(r->headers_in, "Cookie"))
-			== NULL)
-		apr_table_add(r->headers_in, "Cookie", headerString);
-	else
-		apr_table_set(r->headers_in, "Cookie",
-				(apr_pstrcat(r->pool, headerString, ";", currentCookies, NULL)));
-
 	/* do some logging */
 	oidc_debug(r, "adding outgoing header: Set-Cookie: %s", headerString);
 }
@@ -762,7 +857,8 @@ char *oidc_util_get_cookie(request_rec *r, const char *cookieName) {
 				cookie++;
 
 			/* see if we've found the cookie that we're looking for */
-			if ((strncmp(cookie, cookieName, strlen(cookieName)) == 0)  && (cookie[strlen(cookieName)] == '=')) {
+			if ((strncmp(cookie, cookieName, strlen(cookieName)) == 0)
+					&& (cookie[strlen(cookieName)] == '=')) {
 
 				/* skip to the meat of the parameter (the value after the '=') */
 				cookie += (strlen(cookieName) + 1);
@@ -778,11 +874,76 @@ char *oidc_util_get_cookie(request_rec *r, const char *cookieName) {
 	}
 
 	/* log what we've found */
-	oidc_debug(r, "returning \"%s\" = %s", cookieName, rv ? apr_psprintf(r->pool, "\"%s\"", rv) : "<null>");
+	oidc_debug(r, "returning \"%s\" = %s", cookieName,
+			rv ? apr_psprintf(r->pool, "\"%s\"", rv) : "<null>");
 
 	return rv;
 }
 
+#define OIDC_COOKIE_CHUNKS_SEPARATOR "_"
+#define OIDC_COOKIE_CHUNKS_POSTFIX "chunks"
+
+/*
+ * get a cookie value that is split over a number of chunked cookies
+ */
+char *oidc_util_get_chunked_cookie(request_rec *r, const char *cookieName,
+		int chunkSize) {
+	char *cookieValue = NULL;
+	int i = 0;
+	if (chunkSize == 0) {
+		cookieValue = oidc_util_get_cookie(r, cookieName);
+	} else {
+		char *chunkCountName = apr_psprintf(r->pool, "%s%s%s", cookieName,
+				OIDC_COOKIE_CHUNKS_SEPARATOR,
+				OIDC_COOKIE_CHUNKS_POSTFIX);
+		char* chunkCountValue = oidc_util_get_cookie(r, chunkCountName);
+		if (chunkCountValue != NULL) {
+			cookieValue = NULL;
+			char *endptr = NULL;
+			long chunkCount = strtol(chunkCountValue, &endptr, 10);
+			if ((*chunkCountValue != '\0') && (*endptr == '\0')) {
+				cookieValue = "";
+				for (i = 0; i < chunkCount; i++) {
+					char *chunkName = apr_psprintf(r->pool, "%s%s%d",
+							cookieName, OIDC_COOKIE_CHUNKS_SEPARATOR, i);
+					char *chunkValue = oidc_util_get_cookie(r, chunkName);
+					cookieValue = apr_psprintf(r->pool, "%s%s", cookieValue,
+							chunkValue);
+				}
+			}
+		} else {
+			cookieValue = oidc_util_get_cookie(r, cookieName);
+		}
+	}
+	return cookieValue;
+}
+
+/*
+ * set a cookie value that is split over a number of chunked cookies
+ */
+void oidc_util_set_chunked_cookie(request_rec *r, const char *cookieName,
+		const char *cookieValue, apr_time_t expires, int chunkSize) {
+	int i = 0;
+	int cookieLength = strlen(cookieValue);
+	if ((chunkSize == 0) || (cookieLength < chunkSize)) {
+		oidc_util_set_cookie(r, cookieName, cookieValue, expires);
+	} else {
+		int chunkCountValue = cookieLength / chunkSize + 1;
+		const char *ptr = cookieValue;
+		for (i = 0; i < chunkCountValue; i++) {
+			char *chunkName = apr_psprintf(r->pool, "%s%s%d", cookieName,
+					OIDC_COOKIE_CHUNKS_SEPARATOR, i);
+			char *chunkValue = apr_pstrndup(r->pool, ptr, chunkSize);
+			ptr += chunkSize;
+			oidc_util_set_cookie(r, chunkName, chunkValue, expires);
+		};
+		char *chunkCountName = apr_psprintf(r->pool, "%s%s%s", cookieName,
+				OIDC_COOKIE_CHUNKS_SEPARATOR, OIDC_COOKIE_CHUNKS_POSTFIX);
+		oidc_util_set_cookie(r, chunkCountName,
+				apr_psprintf(r->pool, "%d", chunkCountValue), expires);
+	}
+}
+
 /*
  * normalize a string for use as an HTTP Header Name.  Any invalid
  * characters (per http://tools.ietf.org/html/rfc2616#section-4.2 and
@@ -872,7 +1033,8 @@ static apr_byte_t oidc_util_json_string_print(request_rec *r, json_t *result,
 	json_t *value = json_object_get(result, key);
 	if (value != NULL && !json_is_null(value)) {
 		char *s_value = json_dumps(value, JSON_ENCODE_ANY);
-		oidc_error(r, "%s: response contained an \"%s\" entry with value: \"%s\"",
+		oidc_error(r,
+				"%s: response contained an \"%s\" entry with value: \"%s\"",
 				log, key, s_value);
 		free(s_value);
 		return TRUE;
@@ -894,10 +1056,9 @@ static apr_byte_t oidc_util_check_json_error(request_rec *r, json_t *json) {
 }
 
 /*
- * decode a JSON string, check for "error" results and printout
+ * parse a JSON object
  */
-apr_byte_t oidc_util_decode_json_and_check_error(request_rec *r,
-		const char *str, json_t **json) {
+apr_byte_t oidc_util_decode_json_object(request_rec *r, const char *str, json_t **json) {
 
 	json_error_t json_error;
 	*json = json_loads(str, 0, &json_error);
@@ -905,7 +1066,7 @@ apr_byte_t oidc_util_decode_json_and_check_error(request_rec *r,
 	/* decode the JSON contents of the buffer */
 	if (*json == NULL) {
 		/* something went wrong */
-		oidc_error(r, "JSON parsing returned an error: %s", json_error.text);
+		oidc_error(r, "JSON parsing returned an error: %s (%s)", json_error.text, str);
 		return FALSE;
 	}
 
@@ -917,6 +1078,18 @@ apr_byte_t oidc_util_decode_json_and_check_error(request_rec *r,
 		return FALSE;
 	}
 
+	return TRUE;
+}
+
+/*
+ * decode a JSON string, check for "error" results and printout
+ */
+apr_byte_t oidc_util_decode_json_and_check_error(request_rec *r,
+		const char *str, json_t **json) {
+
+	if (oidc_util_decode_json_object(r, str, json) == FALSE)
+		return FALSE;
+
 	// see if it is not an error response somehow
 	if (oidc_util_check_json_error(r, *json) == TRUE) {
 		json_decref(*json);
@@ -955,22 +1128,22 @@ int oidc_util_html_send(request_rec *r, const char *title,
 
 	char *html =
 			"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01//EN\" \"http://www.w3.org/TR/html4/strict.dtd\">\n"
-					"<html>\n"
-					"  <head>\n"
-					"    <meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\">\n"
-					"    <title>%s</title>\n"
-					"    %s\n"
-					"  </head>\n"
-					"  <body%s>\n"
-					"%s\n"
-					"  </body>\n"
-					"</html>\n";
+			"<html>\n"
+			"  <head>\n"
+			"    <meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\">\n"
+			"    <title>%s</title>\n"
+			"    %s\n"
+			"  </head>\n"
+			"  <body%s>\n"
+			"%s\n"
+			"  </body>\n"
+			"</html>\n";
 
 	html = apr_psprintf(r->pool, html,
 			title ? oidc_util_html_escape(r->pool, title) : "",
-			html_head ? html_head : "",
-			on_load ? apr_psprintf(r->pool, " onload=\"%s()\"", on_load) : "",
-			html_body ? html_body : "<p></p>");
+					html_head ? html_head : "",
+							on_load ? apr_psprintf(r->pool, " onload=\"%s()\"", on_load) : "",
+									html_body ? html_body : "<p></p>");
 
 	return oidc_util_http_send(r, html, strlen(html), "text/html", status_code);
 }
@@ -1023,28 +1196,45 @@ int oidc_util_html_send_error(request_rec *r, const char *html_template,
 /*
  * read all bytes from the HTTP request
  */
-static apr_byte_t oidc_util_read(request_rec *r, const char **rbuf) {
+static apr_byte_t oidc_util_read(request_rec *r, char **rbuf) {
+	apr_size_t bytes_read;
+	apr_size_t bytes_left;
+	apr_size_t len;
+	long read_length;
 
-	if (ap_setup_client_block(r, REQUEST_CHUNKED_ERROR) != OK)
+	if (ap_setup_client_block(r, REQUEST_CHUNKED_DECHUNK) != OK)
 		return FALSE;
 
-	if (ap_should_client_block(r)) {
+	len = ap_should_client_block(r) ? r->remaining : 0;
 
-		char argsbuffer[HUGE_STRING_LEN];
-		int rsize, len_read, rpos = 0;
-		long length = r->remaining;
-		*rbuf = apr_pcalloc(r->pool, length + 1);
+	if (len > OIDC_MAX_POST_DATA_LEN) {
+		oidc_error(r, "POST parameter value is too large: %lu bytes (max=%d)",
+				(unsigned long ) len, OIDC_MAX_POST_DATA_LEN);
+		return FALSE;
+	}
 
-		while ((len_read = ap_get_client_block(r, argsbuffer,
-				sizeof(argsbuffer))) > 0) {
-			if ((rpos + len_read) > length) {
-				rsize = length - rpos;
-			} else {
-				rsize = len_read;
-			}
-			memcpy((char*) *rbuf + rpos, argsbuffer, rsize);
-			rpos += rsize;
+	*rbuf = (char *) apr_palloc(r->pool, len + 1);
+	if (*rbuf == NULL) {
+		oidc_error(r, "could not allocate memory for %lu bytes of POST data.",
+				(unsigned long )len);
+		return FALSE;
+	}
+	(*rbuf)[len] = '\0';
+
+	bytes_read = 0;
+	bytes_left = len;
+	while (bytes_left > 0) {
+		read_length = ap_get_client_block(r, &(*rbuf)[bytes_read], bytes_left);
+		if (read_length == 0) {
+			(*rbuf)[bytes_read] = '\0';
+			break;
+		} else if (read_length < 0) {
+			ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
+					"Failed to read POST data from client.");
+			return FALSE;
 		}
+		bytes_read += read_length;
+		bytes_left -= read_length;
 	}
 
 	return TRUE;
@@ -1054,7 +1244,7 @@ static apr_byte_t oidc_util_read(request_rec *r, const char **rbuf) {
  * read form-encoded parameters from a string in to a table
  */
 apr_byte_t oidc_util_read_form_encoded_params(request_rec *r,
-		apr_table_t *table, const char *data) {
+		apr_table_t *table, char *data) {
 	const char *key, *val, *p = data;
 
 	while (p && *p && (val = ap_getword(r->pool, &p, '&'))) {
@@ -1064,8 +1254,8 @@ apr_byte_t oidc_util_read_form_encoded_params(request_rec *r,
 		apr_table_set(table, key, val);
 	}
 
-	oidc_debug(r, "parsed: \"%s\" in to %d elements", data,
-			apr_table_elts(table)->nelts);
+	oidc_debug(r, "parsed: %lu bytes in to %d elements",
+			data ? (unsigned long )strlen(data) : 0 , apr_table_elts(table)->nelts);
 
 	return TRUE;
 }
@@ -1074,7 +1264,7 @@ apr_byte_t oidc_util_read_form_encoded_params(request_rec *r,
  * read the POST parameters in to a table
  */
 apr_byte_t oidc_util_read_post_params(request_rec *r, apr_table_t *table) {
-	const char *data = NULL;
+	char *data = NULL;
 
 	if (r->method_number != M_POST)
 		return FALSE;
@@ -1088,14 +1278,16 @@ apr_byte_t oidc_util_read_post_params(request_rec *r, apr_table_t *table) {
 /*
  * read a file from a path on disk
  */
-apr_byte_t oidc_util_file_read(request_rec *r, const char *path, apr_pool_t *pool, char **result) {
+apr_byte_t oidc_util_file_read(request_rec *r, const char *path,
+		apr_pool_t *pool, char **result) {
 	apr_file_t *fd = NULL;
 	apr_status_t rc = APR_SUCCESS;
 	char s_err[128];
 	apr_finfo_t finfo;
 
 	/* open the file if it exists */
-	if ((rc = apr_file_open(&fd, path, APR_FOPEN_READ | APR_FOPEN_BUFFERED, APR_OS_DEFAULT, r->pool)) != APR_SUCCESS) {
+	if ((rc = apr_file_open(&fd, path, APR_FOPEN_READ | APR_FOPEN_BUFFERED,
+			APR_OS_DEFAULT, r->pool)) != APR_SUCCESS) {
 		oidc_warn(r, "no file found at: \"%s\"", path);
 		return FALSE;
 	}
@@ -1198,9 +1390,9 @@ apr_byte_t oidc_util_json_array_has_value(request_rec *r, json_t *haystack,
 		}
 	}
 
-//	oidc_debug(r,
-//			"returning (%d=%d)", i,
-//			haystack->value.array->nelts);
+	//	oidc_debug(r,
+	//			"returning (%d=%d)", i,
+	//			haystack->value.array->nelts);
 
 	return (i == json_array_size(haystack)) ? FALSE : TRUE;
 }
@@ -1208,7 +1400,8 @@ apr_byte_t oidc_util_json_array_has_value(request_rec *r, json_t *haystack,
 /*
  * set a HTTP header
  */
-void oidc_util_set_header(request_rec *r, const char *s_name, const char *s_value) {
+void oidc_util_set_header(request_rec *r, const char *s_name,
+		const char *s_value) {
 	/*
 	 * sanitize the header value by replacing line feeds with spaces
 	 * just like the Apache header input algorithms do for incoming headers
@@ -1304,8 +1497,7 @@ void oidc_util_set_app_infos(request_rec *r, const json_t *j_attrs,
 
 		} else if (json_is_integer(j_value)) {
 
-			if (sprintf(s_int, "%ld",
-					(long)json_integer_value(j_value)) > 0) {
+			if (sprintf(s_int, "%ld", (long) json_integer_value(j_value)) > 0) {
 				/* set long value in the application header whose name is based on the key and the prefix */
 				oidc_util_set_app_info(r, s_key, s_int, claim_prefix, as_header,
 						as_env_var);
@@ -1441,8 +1633,8 @@ apr_byte_t oidc_util_spaced_string_equals(apr_pool_t *pool, const char *a,
 /*
  * see if a particular value is part of a space separated value
  */
-apr_byte_t oidc_util_spaced_string_contains(apr_pool_t *pool,
-		const char *str, const char *match) {
+apr_byte_t oidc_util_spaced_string_contains(apr_pool_t *pool, const char *str,
+		const char *match) {
 	apr_hash_t *ht = oidc_util_spaced_string_to_hashtable(pool, str);
 	return (apr_hash_get(ht, match, APR_HASH_KEY_STRING) != NULL);
 }
@@ -1480,16 +1672,17 @@ apr_byte_t oidc_json_object_get_int(apr_pool_t *pool, json_t *json,
 /*
  * merge two JSON objects
  */
- apr_byte_t oidc_util_json_merge(json_t *src, json_t *dst) {
+apr_byte_t oidc_util_json_merge(json_t *src, json_t *dst) {
 
 	const char *key;
 	json_t *value = NULL;
 	void *iter = NULL;
 
-	if ((src == NULL) || (dst == NULL)) return FALSE;
+	if ((src == NULL) || (dst == NULL))
+		return FALSE;
 
 	iter = json_object_iter(src);
-	while(iter) {
+	while (iter) {
 		key = json_object_iter_key(iter);
 		value = json_object_iter_value(iter);
 		json_object_set(dst, key, value);
@@ -1517,36 +1710,56 @@ void oidc_util_table_add_query_encoded_params(apr_pool_t *pool,
 }
 
 /*
- * merge provided keys and client secret in to a single hashtable
+ * create a symmetric key from a client_secret
  */
-apr_hash_t * oidc_util_merge_symmetric_key(apr_pool_t *pool, apr_hash_t *keys,
-		const char *client_secret, const char *hash_algo) {
-	apr_jwt_error_t err;
-	apr_jwk_t *jwk = NULL;
+apr_byte_t oidc_util_create_symmetric_key(request_rec *r,
+		const char *client_secret, int r_key_len, const char *hash_algo,
+		apr_byte_t set_kid, oidc_jwk_t **jwk) {
+	oidc_jose_error_t err;
 	unsigned char *key = NULL;
 	unsigned int key_len;
-	apr_hash_t *result = NULL;
 
-	result = (keys != NULL) ? apr_hash_copy(pool, keys) : apr_hash_make(pool);
-
-	if (client_secret != NULL) {
+	if ((client_secret != NULL) && (strlen(client_secret) > 0)) {
 
 		if (hash_algo == NULL) {
 			key = (unsigned char *) client_secret;
 			key_len = strlen(client_secret);
 		} else {
 			/* hash the client_secret first, this is OpenID Connect specific */
-			apr_jws_hash_bytes(pool, hash_algo,
+			oidc_jose_hash_bytes(r->pool, hash_algo,
 					(const unsigned char *) client_secret,
 					strlen(client_secret), &key, &key_len, &err);
 		}
 
-		if (apr_jwk_parse_symmetric_key(pool, NULL, key, key_len, &jwk,
-				&err) == TRUE) {
-			apr_hash_set(result, jwk->kid, APR_HASH_KEY_STRING, jwk);
+		if ((key != NULL) && (key_len > 0)) {
+			if ((r_key_len != 0) && (key_len >= r_key_len))
+				key_len = r_key_len;
+			oidc_debug(r, "key_len=%d", key_len);
+			*jwk = oidc_jwk_create_symmetric_key(r->pool, NULL, key, key_len,
+					set_kid, &err);
+		}
+
+		if (*jwk == NULL) {
+			oidc_error(r,
+					"could not create JWK from the provided secret %s: %s",
+					client_secret, oidc_jose_e2s(r->pool, err));
+			return FALSE;
 		}
 	}
 
+	return TRUE;
+}
+
+/*
+ * merge provided keys and client secret in to a single hashtable
+ */
+apr_hash_t * oidc_util_merge_symmetric_key(apr_pool_t *pool, apr_hash_t *keys,
+		oidc_jwk_t *jwk) {
+	apr_hash_t *result =
+			(keys != NULL) ? apr_hash_copy(pool, keys) : apr_hash_make(pool);
+	if (jwk != NULL) {
+		apr_hash_set(result, jwk->kid, APR_HASH_KEY_STRING, jwk);
+	}
 	return result;
 }
 
@@ -1555,21 +1768,19 @@ apr_hash_t * oidc_util_merge_symmetric_key(apr_pool_t *pool, apr_hash_t *keys,
  */
 apr_byte_t oidc_util_hash_string_and_base64url_encode(request_rec *r,
 		const char *openssl_hash_algo, const char *input, char **output) {
-	apr_jwt_error_t err;
+	oidc_jose_error_t err;
 	unsigned char *hashed = NULL;
 	unsigned int hashed_len = 0;
-	if (apr_jws_hash_bytes(r->pool, openssl_hash_algo,
+	if (oidc_jose_hash_bytes(r->pool, openssl_hash_algo,
 			(const unsigned char *) input, strlen(input), &hashed, &hashed_len,
 			&err) == FALSE) {
-		oidc_error(r,
-				"apr_jws_hash_bytes returned an error: %s", err.text);
+		oidc_error(r, "oidc_jose_hash_bytes returned an error: %s", err.text);
 		return FALSE;
 	}
 
 	if (oidc_base64url_encode(r, output, (const char *) hashed, hashed_len,
 			TRUE) <= 0) {
-		oidc_error(r,
-				"oidc_base64url_encode returned an error: %s", err.text);
+		oidc_error(r, "oidc_base64url_encode returned an error: %s", err.text);
 		return FALSE;
 	}
 	return TRUE;
